import logging

from django.conf import settings

from rest_framework import status
from rest_framework.response import Response

from generic.exceptions import ObjectNotFoundError, ValidationError
from generic.logger import get_log_msg

LOGGER = logging.getLogger('views')


def handle_error(error, status_code):
    """
    General error handler with error message and http error code being passed.
    """
    content = {settings.ERROR_KEY: str(error)}
    return Response(content, status=status_code)


class GenericViewset:
    """
    Provides generic functionality for http post, get, patch and delete calls.
    """
    serializer_class = None
    service = None
    request = None

    def __init__(self, serializer_class, service, request):
        self.serializer_class = serializer_class
        self.service = service
        self.request = request

    def create(self, args_validator=None):
        """
        Creates and persists passed object if valid.
        """
        LOGGER.info(get_log_msg(self.request, self.request.user))
        data = self.request.data

        # Validate arguments
        try:
            if args_validator is not None:
                args_validator(data)
        except ValidationError as error:
            return handle_error(error, status.HTTP_400_BAD_REQUEST)

        # Apply serializer
        serializer = self.serializer_class(data=data, many=False)
        if serializer.is_valid():
            return self._create_instance(**serializer.data)

        return handle_error('Validation failed.', status.HTTP_400_BAD_REQUEST)

    def _create_instance(self, **data):
        try:
            instance = self.service(requesting_user=self.request.user, **data)
            serializer = self.serializer_class(instance=instance, many=False)
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as error:
            return handle_error(error, status.HTTP_400_BAD_REQUEST)
        except ObjectNotFoundError as error:
            return handle_error(error, status.HTTP_404_NOT_FOUND)

    def list(self):
        """
        Retrieves all objects returned by the provided service.
        """
        LOGGER.info(get_log_msg(self.request, self.request.user))
        return self._list_instances()

    def _list_instances(self):
        instances = self.service(requesting_user=self.request.user)
        serializer = self.serializer_class(instance=instances, many=True)
        return Response(serializer.data, status=status.HTTP_200_OK)

    def retrieve(self, key):
        """
        Retrieves a single object by its primary key.
        """
        LOGGER.info(get_log_msg(self.request, self.request.user))
        return self._retrieve_instance(key)

    def _retrieve_instance(self, key):
        try:
            instance = self.service(key, requesting_user=self.request.user)
            serializer = self.serializer_class(instance=instance, many=False)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except ObjectNotFoundError as error:
            return handle_error(error, status.HTTP_404_NOT_FOUND)

    def destroy(self, key):
        """
        Deletes a single object by its primary key.
        """
        LOGGER.info(get_log_msg(self.request, self.request.user))
        return self._destroy_instance(key)

    def _destroy_instance(self, key):
        try:
            instance = self.service(key, requesting_user=self.request.user)
            serializer = self.serializer_class(instance=instance, many=False)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except ObjectNotFoundError as error:
            return handle_error(error, status.HTTP_404_NOT_FOUND)

    def partial_update(self, key, args_validator=None):
        """
        Partially updates an existing object without overwriting it.
        """
        LOGGER.info(get_log_msg(self.request, self.request.user))
        data = self.request.data

        # Validate arguments
        try:
            if args_validator is not None:
                args_validator(data)
        except ValidationError as error:
            return handle_error(error, status.HTTP_400_BAD_REQUEST)

        # Apply serializer
        serializer = self.serializer_class(data=data, many=False)
        if serializer.is_valid():
            return self._update_instance(key, **serializer.data)

        return handle_error('Validation failed.', status.HTTP_400_BAD_REQUEST)

    def _update_instance(self, key, **data):
        try:
            instance = self.service(key, requesting_user=self.request.user, **data)
            serializer = self.serializer_class(instance=instance, many=False)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except ObjectNotFoundError as error:
            return handle_error(error, status.HTTP_404_NOT_FOUND)
        except ValidationError as error:
            return handle_error(error, status.HTTP_400_BAD_REQUEST)



from django.core.exceptions import ImproperlyConfigured
from django.db.models.query import QuerySet

from generic.exceptions import ValidationError


def unique_field(repository, field):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that field is unique.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if not repository.is_unique(instance, field):
                raise ValidationError('This value has already been used.',
                                      field=field)
            return True

    return _checker


def unique_together_fields(repository, *fields):
    """
    :param repository: Repository that should be used to perform validation
    :param fields: Fields to be validated
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that the combination of fields is unique.
    Checks if a field is unique.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if len(repository.search_by(instance, *fields)) is not 0:
                raise ValidationError('These values have already been used.',
                                      field=fields)
            return True

    return _checker


def required_field(repository, field):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that field is present and not None.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.is_none(instance, field):
                raise ValidationError('Mandatory value was not provided.',
                                      field=field)
            return True

    return _checker


def max_length(repository, field, length):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param length: Maximal length for field
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that the string in field comprises at most length characters.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.is_too_long(instance, field, length):
                raise ValidationError(("Maximal length of {0} chars "
                                       "exceeded.").format(length),
                                      field=field)
            return True

    return _checker


def min_length(repository, field, length):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param length: Maximal length for field
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that the string in field comprises at least length characters.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.is_too_short(instance, field, length):
                raise ValidationError(("Minimal length of {0} characters "
                                       "not met.").format(length),
                                      field=field)

            return True

    return _checker


def read_only_field(repository, field):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that read only fields are not written to.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.contains_arg(instance, field):
                raise ValidationError('Attempt to write read only field.',
                                      field=field)
            return True

    return _checker


def min_char_classes(repository, field, char_classes):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param char_classes: Number of different character classes need to be
    present; can be lower/upper case letters, digits or punctuation
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that the string in field comprises a sufficient amount of
    different character classes.
    """
    if char_classes < 0 or char_classes > 4:
        raise ImproperlyConfigured(("Number of character classes "
                                    "must be  in range 0-4."))

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if not repository.contains_char_classes(instance, field, char_classes):
                raise ValidationError(("Field must contain {0} of the "
                                       "following characters: upper case letters, "
                                       "lower case letters, digits and special "
                                       "characters.").format(char_classes),
                                      field='password')

            return True

    return _checker


def email_field(repository, field):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that an email address field is populated with a valid email
    address.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if not repository.is_valid_email(instance, field):
                raise ValidationError('This email address is invalid.', field='email')
            return True

    return _checker


def image_dim_validator(repository, field, width, height):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param width: Maximum image width
    :param height: Maximum image height
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that an image is not too tall or wide.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.image_too_tall(instance, field, height):
                msg = "Image is too tall: Maximum height is {0}.".format(height)
                raise ValidationError(msg, field=field)
            if repository.image_too_wide(instance, field, width):
                msg = "Image is too wide: Maximum width is {0}.".format(width)
                raise ValidationError(msg, field=field)
            return True

    return _checker


def image_size_validator(repository, field, size):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param size: Maximum image size in bytes
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that an image is not too large.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.image_too_large(instance, field, size):
                msg = "Image is too large: Maximum size is {}B.".format(size)
                raise ValidationError(msg, field=field)
            return True

    return _checker


def image_square_validator(repository, field):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :return: ValidationError if the validation failed

    Ensures that image is quadratic.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if not repository.is_square(instance, field):
                msg = 'Image must be quadratic.'
                raise ValidationError(msg, field=field)
            return True

    return _checker


def validate_many(*validators):
    """
    :param validators: List of validators to be applied

    Allows to apply multiple validators on one instance.
    """

    def _checker(instance):
        return all(validator(instance) for validator in validators)

    return _checker


def file_size_validator(repository, field, size):
    """
    :param repository: Repository that should be used to perform validation
    :param field: Field to be validated
    :param size: Maximum image size in bytes
    :return: True is validation was successful
    :raises: ValidationError if the validation failed

    Ensures that a file is not too large.
    """

    def _checker(instance):
        if isinstance(instance, QuerySet):
            return all(_checker(i) for i in instance)
        else:
            if repository.file_too_large(instance, field, size):
                msg = "File is too large: Maximum size is {}B.".format(size)
                raise ValidationError(msg, field=field)
            return True

    return _checker



class ValidationError(Exception):
    """
    Error that is raised while validating data.
    """

    def __init__(self, message, field=None, code=None, params=None):
        from django.utils.encoding import force_text
        """
        ValidationError can be passed any object that can be printed (usually
        a string), a list of objects or a dictionary.
        """

        self.field = field
        self.code = code
        self.params = params
        self.msg = force_text(message)
        self.messages = [self.msg]

    def __str__(self):
        return "{0} ({1})".format(self.msg, self.field)


class ObjectNotFoundError(Exception):
    """
    Error that is raised while getting object.
    """

    def __init__(self, message, modelclass=None, code=None, params=None):
        from django.utils.encoding import force_text
        """
        ObjectNotFoundError can be passed any object that can be printed (usually
        a string), a list of objects or a dictionary.
        """

        self.modelclass = modelclass
        self.code = code
        self.params = params
        self.msg = force_text(message)
        self.messages = [self.msg]

    def __str__(self):
        try:
            msg = "{0} ({1})".format(self.msg, self.modelclass.__name__)
        except AttributeError:
            msg = "{0} ({1})".format(self.msg, self.modelclass)
        return msg


class ExtLibError(Exception):
    """
    Error that is raised while calling external lib.
    """

    def __init__(self, message, lib=None, code=None, params=None):
        from django.utils.encoding import force_text
        """
        ExtLibError can be passed any object that can be printed (usually
        a string), a list of objects or a dictionary.
        """

        self.lib = lib
        self.code = code
        self.params = params
        self.msg = force_text(message)
        self.messages = [self.msg]

    def __str__(self):
        return "{0} ({1})".format(self.msg, self.lib)


class InternalError(Exception):
    """An unexpected general error"""

    def __init__(self, message, fct=None, code=None, params=None):
        from django.utils.encoding import force_text
        """
        InternalError can be passed any object that can be printed (usually
        a string), a list of objects or a dictionary.
        """

        self.fct = fct
        self.code = code
        self.params = params
        self.msg = force_text(message)
        self.messages = [self.msg]

    def __str__(self):
        return "{0} ({1})".format(self.msg, self.fct)



from django.contrib.auth.hashers import make_password
from rest_framework.utils import model_meta

from generic.repositories import GenericRepository


def generic_retrieve_single_service(repository):
    """
    :param repository: Repository that will be called
    :return: Instance matching pk

    Returns an instance by pk.
    """

    def _retrieve(pk, requesting_user):
        instance = repository.get_by_id(pk, requesting_user)
        return instance

    return _retrieve


def generic_retrieve_all_service(repository):
    """
    :param repository: Repository that will be called
    :return: Queryset with all stored instances

    Returns all stored instances.
    """

    def _retrieve(requesting_user=None):
        instances = repository.list_all()
        return instances

    return _retrieve


def generic_retrieve_all_by_owner_service(repository):
    """
    :param repository: Repository that will be called
    :return: Queryset with all owned instances

    Returns all stored instances owned by user.
    """

    def _retrieve(requesting_user):
        instances = repository.find_by_user(requesting_user)
        return instances

    return _retrieve


def generic_delete_service(repository):
    """
    :param repository: Repository that will be called
    :return: Instance that was deleted

    Deletes an instance by pk.
    """

    def _delete(pk, requesting_user):
        instance = repository.get_by_id(pk, requesting_user)
        repository.delete_by_id(pk, requesting_user)
        return instance

    return _delete


def generic_update_service(repository, validator):
    """
    :param repository: Repository that will be called
    :param validator: Instance validator
    :return: Updated and saved instance

    Updates an existing instance.
    """

    def _update(pk, requesting_user, **kwargs):

        # Get instance and meta info
        info = model_meta.get_field_info(repository.model_class)
        instance = repository.get_by_id(pk, requesting_user)

        # One-to-many relations
        for field_name, relation_info in info.relations.items():
            if (not relation_info.to_many
                    and field_name in kwargs.keys()
                    and kwargs[field_name] is not None):
                model = relation_info.related_model

                pk = kwargs[field_name]
                repo = GenericRepository(model)
                related_instance = repo.get_by_id(pk)

                setattr(instance, field_name, related_instance)

        # Set regular fields and validate
        for field_name, field_info in info.fields.items():
            if field_name in kwargs.keys():
                setattr(instance, field_name, kwargs[field_name])

        validator(instance)

        # Set hashed password
        if hasattr(repository.model_class, 'password') and 'password' in kwargs:
            instance.password = make_password(kwargs['password'])

        # Save and return
        repository.persist(instance)
        return instance

    return _update


def generic_create_service(repository, validator, factory):
    """
    :param repository: Repository that will be called
    :param validator: Instance validator
    :param factory: Factory that creates the instance
    :return: Created and saved instance

    Returns and persists a new instance.
    """

    def _create(requesting_user, **kwargs):

        # Get owner and meta info
        info = model_meta.get_field_info(repository.model_class)
        if hasattr(repository.model_class, 'owner'):
            kwargs['owner'] = requesting_user.id

        # One-to-many relations
        for field_name, relation_info in info.relations.items():
            if (not relation_info.to_many
                    and field_name in kwargs.keys()
                    and kwargs[field_name] is not None):
                model = relation_info.related_model

                pk = kwargs[field_name]
                repo = GenericRepository(model)
                related_instance = repo.get_by_id(pk)

                kwargs[field_name] = related_instance

        # Create instance and validate
        instance = factory(repository.model_class, **kwargs)
        validator(instance)

        # Set hashed password
        if hasattr(repository.model_class, 'password') and 'password' in kwargs:
            instance.password = make_password(kwargs['password'])

        # Save and return
        repository.persist(instance)
        return instance

    return _create


def store_image_service(repository, validator, field):
    """
    :param repository: Repository that will be called
    :param validator: Instance validator
    :param field: Name of ImageField
    :return: instance with image

    Adds image to instance. Also stores image on disk.
    """

    def _store(instance, image):
        instance = repository.store_image(instance, image, field)
        validator(instance)
        instance.save()
        return instance

    return _store


def remove_image_service(repository, field):
    """
    :param repository: Repository that will be called
    :param field: Name of ImageField
    :return: User instance without image

    Removes image from instance. Also deletes image from disk.
    """

    def _remove(pk, requesting_user):
        instance = repository.get_by_id(pk, requesting_user)
        image = getattr(instance, field)
        image.delete(save=True)
        return instance

    return _remove


def store_file_service(repository, validator, field):
    """
    :param repository: Repository that will be called
    :param validator: Instance validator
    :param field: Name of FileField
    :return: instance with file

    Adds file to instance. Also stores file on disk.
    """

    def _store(instance, file):
        instance = repository.store_file(instance, file, field)
        validator(instance)
        instance.save()
        return instance

    return _store


def remove_file_service(repository, field):
    """
    :param repository: Repository that will be called
    :param field: Name of FileField
    :return: User instance without file

    Removes file from instance. Also deletes file from disk.
    """

    def _remove(pk, requesting_user):
        instance = repository.get_by_id(pk, requesting_user)
        file = getattr(instance, field)
        file.delete(save=True)
        return instance

    return _remove



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-generic-rest',
    version='0.4.5',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='Provides reusable REST API view components.',
    long_description=README,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django',
        'djangorestframework'
    ],
)
